/* Originally develeoped/debugged by                                         */
/* Bob Blair      PATROL Development      BMC Software, Austin               */
/*                bblair@bmc.com (only send thanks, no complaints please)    */
/* Distributed and updated as necessary by                                   */
/* Brian Carr     brian_carr@usa.net (comments and suggestions welcome)      */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <io.h>
#include <process.h>
#include <errno.h>
#include <os2.h>      /* for pipe & waitpid */
/* #include <os2/os2pipe.h>      |* for pipe & waitpid */
#include "pipe.h"      /* for pipe & waitpid */
/* #include <sys/wait.h>                                                     */
/* #include <malloc/malloc.h>                                                */
#include "popen.h"
struct _popids * popidhead = NULL;

int
postore(FILE *f, int pid )
{
   struct _popids *p, *q;

   p = ( struct _popids * ) malloc( sizeof( struct _popids ) );
   if (p) {
         p->next = NULL;
         p->f = f;
         p->pid = pid;
         if (!popidhead)
            popidhead = p;
         else {
            for( q = popidhead; q->next; q = q->next ) { ;} ;
            q->next = p;
         }
         return(0);
   } else
      return(-1);
}

int
pounstore(FILE *f)
{
   struct _popids *p, *q;
   int pid;

   if (popidhead) {
      for( q = NULL, p = popidhead;
           p->next && p->f != f;
           p = p->next) { ; };
      if (p->f == f) {
         if (q == NULL)
            popidhead = p->next;
         else
            q->next = p->next;
         pid = p->pid;
         free(p);
      } else
         pid = 0;
   } else
     pid = 0;
   return(pid);
}
void restore (int org_handle, int org_private, int handle)
{
  int saved_errno;

  saved_errno = errno;
  close (handle);
  dup (org_handle);
  close (org_handle);
  errno = saved_errno;
}


FILE * backout( unsigned long *ph, int org_handle, int org_private, int handle)
{
      int saved_errno;

      restore(org_handle, org_private, handle);
      saved_errno = errno;
      close (ph[0]);
      close (ph[1]);
      errno = saved_errno;
      return(NULL);
}

FILE *
popen (const char *command, const char *mode)
{
  unsigned long ph[2];
  int  pipe_local, pipe_remote, handle, i, org_handle, org_private = 0;
  FILE *f;
/*  int saved_errno;                                                         */
  char *sh;
/*  char *opt;                                                               */

  if (mode[0] != 'r' && mode[0] != 'w')
    {
      errno = EINVAL;
      return( ( FILE * ) NULL );
    }
  if( pipe( ph ) == -1)
    return (NULL);
  pipe_local = ph[0];
  pipe_remote = ph[1];
  if (mode[0] == 'r')
    handle = 1;
  else
    handle = 0;

  org_handle = dup (handle);
  if (org_handle == -1)
    return (NULL);
  if (close (handle) == -1)
    return (NULL);
  i = dup (pipe_remote);
  if (i != handle)
    {
      errno = EBADF;
      return( backout( ph, org_handle, org_private, handle ) );
    }
  if (close (pipe_remote) == -1)
      return( backout ( ph, org_handle, org_private, handle ) );
  f = fdopen (pipe_local, mode);
  if (f == NULL)
      return( backout ( ph, org_handle, org_private, handle ) );
  sh = getenv( "COMSPEC" );
  if( sh == NULL )
    {
      fclose (f);
      errno = ENOENT;
      return( backout ( ph, org_handle, org_private, handle ) );
    }
/*opt = "-c";                                                                */
  i = spawnlp (P_NOWAIT, sh, sh, "/C", command, NULL);
  if (i == -1)
    {
      fclose (f);
      f = NULL;
      return( backout( ph, org_handle, org_private, handle ) );
    }
  restore( org_handle, org_private, handle );
  postore( f, i );

  return( f );
}

int
pclose(FILE *f)
{
   int pid, ret;
   pid = pounstore( f );
   if( pid )
      {
         _cwait( &ret, pid, WAIT_CHILD );
         fclose(f);
         return(ret);
      }
   errno = ECHILD;
   return(-1);
}

#ifdef TEST
void main(int argc, char **argv)
{
   char *cmd = "DIR";
   char buf[82];
   FILE *f;
   int rc;

   if (argc > 1)
      cmd = argv[1];

   f = popen(cmd, "r");
   if (!f)
     printf("popen failed. errno = %d", errno);
   else {
     printf("**************** Start of command output *****************\n");
     while (fgets(buf,80,f)) {
        printf("%s",buf);
     } /* endwhile */
     printf("****************   End of command output *****************\n");
     rc = pclose(f);
     printf("Return code from pclose() was %d\n", rc);
   }
}
#endif


