#include <string.h>
#include <stdlib.h>

static char mime_char[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

char *mime_encode_base64 (const char *s, int size)
{
  int mem_size, col, i, shift, linefeed;
  unsigned char bits;
  char *buffer, *ptr;
  if (size <= 0)
    return NULL;
  /* first let's check the memory required. */
  mem_size = (size + 2) / 3 * 4;    /* 8 bit to 6 bit conversion */
  linefeed = (mem_size + 63) / 64;
  buffer = (char *)malloc (mem_size + linefeed + 1);
  if (!buffer)
    return NULL;
  /* initialize */
  buffer[mem_size + linefeed] = '\0';
  ptr = buffer;
  col = 0;
  shift = -2;
  bits = 0;
  for (i = 0; i < size; i++)
    {
      switch (shift)
	{
	  case 0:  /* bits is full */
	    *ptr = mime_char[bits]; ptr++;
	    col++;
	    if (col == 64)
	      {
		*ptr = '\n'; ptr++;
		col = 0;
	      }
	  case -2: /* load new character with fresh bits */
	    bits = *s & 0x03;
	    *ptr = mime_char[*s >> 2]; ptr++;
	    shift = 4;
	    s++;
	    break;
	  default:
	    *ptr = mime_char[(bits << shift) | (*s >> (8 - shift))]; ptr++;
	    bits = *s & (0xff >> shift);
	    shift -= 2;
	    s++;
	}
      col++;
    }
  if (shift)
    *ptr = mime_char[bits << shift];
  else
    *ptr = mime_char[bits];
  ptr++;
  col++;
  col %= 4;
  if (col)
    for (i = 4; i > col; i--)
      {
	*ptr = '='; ptr++;
      }
  *ptr = '\n';
  return buffer;
}
