/************************************************************************\
*                                                                        *
* CookMail 2 - Binary Tree Library                                       *
* (c) Copyright Heng Yuan 1997                                           *
* All Rights Reserved.                                                   *
*                                                                        *
* Do NOT distribute this program in any modified format.                 *
*                                                                        *
* Description:                                                           *
*      A Form Handling CGI script for Unix & Windows NT                  *
*                                                                        *
\************************************************************************/

#ifdef DOS
#include <windows.h>
#endif

#include "btree.h"
#include <string.h>
#include <stdlib.h>

#define newdata(a, b, c, d)  a->prev = NULL; a->next = NULL; \
			     a->data = b; a->value = c; \
                             a->linearnext = NULL; \
                             if (d) d->linearnext = a;

/* local function, should not be exported */
/*

   This tree serves two purposes:
   1. search data, find if the data is found
   2. If the data is found, it gives the location of the data
      in btree;
      If not found, it gives the available location for adding
      in btree;

*/
int searchtree (btree **tree, const char *data)
{
  int value;
  btree *tmp = *tree;
  if (tmp == NULL)  /* no pointer */
    return -1;
  do
    {
      value = strcmp (tmp->data, data);
      if (value == 0)
	break;
      else if (value < 0)
	{
	  if (tmp->prev != NULL)
	    tmp = tmp->prev;
	  else
	    break;
	}
      else if (value > 0)
	{
	  if (tmp->next != NULL)
	    tmp = tmp->next;
	  else
	    break;
	}
    }
  while (value != 0);

  *tree = tmp;
  return value;
}

/* add a new data to the tree */
btree *addtree (btree *tree, char *data, char *datavalue, btree *linear)
{
  btree *tmp;
  int   value;
  if (tree)
    {
      tmp = tree;
      value = searchtree (&tmp, data);
      if (value == 0)  /* found */
	{
	  if (tmp->value)
	    free (tmp->value);        /* replace old value with new one */
	  tmp->value = datavalue;     /* assume datavalue is allocated */
	}
      else
	{
	  btree *tmp2;
	  tmp2 = (btree *)malloc (sizeof (btree));
	  if (tmp2 == NULL)
	    return tree;            /* allocation failed */
	  if (value < 0)
	    tmp->prev = tmp2;
	  else
	    tmp->next = tmp2;
	  tmp = tmp2;
	  newdata (tmp, data, datavalue, linear);
	}
    }
  else
    {
      tmp = (btree *)malloc (sizeof (btree));
      if (tmp == NULL)
	return NULL;
      newdata (tmp, data, datavalue, linear);
    }
  return tmp;
}

/* let's bet our luck, hope it won't overflow */
void trashtree (btree *tree)
{
  if (tree->prev)
    trashtree (tree->prev);
  if (tree->next)
    trashtree (tree->next);
  /* free this tree */
  free (tree->data);
  if (tree->value)
    free (tree->value);
  free (tree);
}

/* is data store in the tree? */
int intree (btree *tree, char *data)
{
  btree *tmp = tree;
  return searchtree (&tmp, data);
}

/* get davavalue from tree */
char *getdata (btree *tree, char* data)
{
  btree *tmp = tree;
  if (searchtree (&tmp, data) == 0)
    return getnodedata (tmp);
  return NULL;
}
