/************************************************************************\
*                                                                        *
* HTML CGI Library - I/O Routines                                        *
* (c) Copyright Heng Yuan 1996                                           *
* All Rights Reserved.                                                   *
*                                                                        *
* Do NOT distribute this program in any modified format.                 *
*                                                                        *
* Description:                                                           *
*      A Form Handling CGI script for Unix                               *
*                                                                        *
\************************************************************************/

#ifdef OS2
#define strcasecmp strcmp
#endif
#ifdef DOS
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "string.h"
#include <stdarg.h>
#include "html.h"

/* some thoughts

  btree is used extensively in version 2 of CookMail to
  store formdata.  Predefined "environmental variables"
  (quoted because this newer version no longer depend on
  the system getenv/putenv for these variables) are also
  stored in btree.  The searches for these predefined
  variables can be optimized by having a btree pointer
  point to them.  This does not add search cost to the
  program at all since addtree () returns the pointer
  of the newly added node.

*NOTE*
  In this version 2, PREVENV is no longer necessary.

*/


/* some exported global variables */
int STARTCGIENV = 0;

/* global binary tree pointer */
btree *FORMDATA = NULL;

/* multiple selection mark */
char MULTIPLEMARK[] = "MULTIPLEMARK";
char defaultmultiplemark[] = ", ";
btree *bt_multiplemark;
#define multiplemark (getnodedata (bt_multiplemark))

/* force concatination of several inputs */
char MULTIPLE[] = "MULTIPLE";
char defaultmultiple[] = "ON";
btree *bt_multiple;
#define multiple (getnodedata (bt_multiple))

/* empty field handling */
/* EMPTYHANDLE valid values are IGNORE, ERROR, NOIGNORE */
char EMPTYHANDLE[] = "EMPTYHANDLE";
char defaultempty[] = "IGNORE";
btree *bt_emptyhandle;
#define emptyhandle (getnodedata (bt_emptyhandle))

char EMPTYERROR[] = "EMPTYERROR";
btree *bt_emptyerror;
#define emptyerror (getnodedata (bt_emptyerror))

/* this env shouldn't be used */
/* version 2:  remove prevenv */
char *prevenv = NULL;

/* some local global variables */
/* signiture used to mark the beginning of the cgi script environment */
static char signature[] = "SHLVL=HENG";
static char *inputstring = NULL;
int inputlength = 0;
int freeinput = 0;

/* for ETC count */
btree *linearnext = NULL;
btree *linearstart = NULL;

/* local functions */
char hex2c (const char *);  /* a very limited hex number to char converter */
int getinput( void );   /* get string from input for GET, POST, or QUERY method */
int transf( void );                   /* translate inputstring into environment */
btree *addtreeinput (btree *tree, char *data, char *datavalue);

#ifndef putenv
int putenv (const char *string);
#endif

void startcgi ()
{
  int i;
  bt_emptyerror = addtreeglobal (strdup (EMPTYERROR), NULL);
  bt_emptyhandle = addtreeglobal (strdup (EMPTYHANDLE), strdup (defaultempty));
  bt_multiple = addtreeglobal (strdup (MULTIPLE), strdup (defaultmultiple));
  bt_multiplemark = addtreeglobal (strdup (MULTIPLEMARK), strdup (defaultmultiplemark));

  for (i = 0;  environ[i]!= 0; i++)
    { ; }
  STARTCGIENV = i;
}

/* this function tells the httpd to display another HTML */
void showhtml (const char *html)
{
  putenv (signature);
  #ifdef DOS
    putsf ("HTTP/1.0 200 OK\n");
  #endif
  putsf ("Location:  %s\n\n", html);
}

/* this function start the cgi output */
void showoutput (CONTENT_TYPES type)
{
  if (getenv ("SHLVL") == NULL)
    {
      putenv (signature);
      #ifdef DOS
        putsf ("HTTP/1.0 200 OK\n");
      #endif
      putsf ("Content-type:  %s\n\n", content_types[type]);
    }
}

/* this function let the parent find both userinput and environ var values */
char *getformdata (const char *data)
{
  btree *tmp = FORMDATA;
  char *str;
  if (searchtree (&tmp, data) == 0)
    {
      if (getnodedata (tmp))
	return getnodedata (tmp);
      return "";
    }
  str = getenv (data);
  if (!str)
    return "";
  return str;
}

/* similar to printf, but with error checking */
int putsf (const char *fmt, ...)
{
  int error;

  va_list args;
  va_start (args, fmt);

  error = vprintf (fmt, args);
  va_end (args);

  /* hmm, break in transfer. Let's terminate */
  if (error < 1)
    exit (1);
  return error;
}

char hex2c (const char *c)
{
  char ch;
  if (*c >= '0' && *c <= '9')
    ch = *c - '0';
  else if (*c >= 'a')
    ch = *c - 'a' + 10;
  else
    ch = *c - 'A' + 10;

  ch *= 16;
  c++;

  if (*c >= '0' && *c <= '9')
    ch += *c - '0';
  else if (*c >= 'a')
    ch += *c - 'a' + 10;
  else
    ch += *c - 'A' + 10;
  return ch;
}

/* get input string from GET, POST, or QUERY methods */
int getinput ()
{
  if (strcmp (REQUEST_METHOD, "POST") == 0)
    {
      inputlength = atoi (CONTENT_LENGTH);
      inputstring = (char *)malloc (inputlength + 1);
      if (inputstring == NULL)
        return -1;
      freeinput = 1;
      return fgets (inputstring, inputlength + 1, stdin) != inputstring;
    }
  else if ((strcmp (REQUEST_METHOD, "GET") == 0) ||
	   (strcmp (REQUEST_METHOD, "QUERY") == 0))
    {
      inputstring = QUERY_STRING;
      inputlength = strlen (inputstring);
    }
  else
    return -1;  /* no such method */
  return 0;
}

/* check if the string is empty */
int isemptyenv (const char *str)
{
  int none, length;
  length = strlen (str);
  none = strspn (str, " \t\n\r");
  if (none == length)
    return 1;
  return 0;
}

/* handle string parsed by transf () */
int handleenv (char *str)
{
  char *data;
  char *datavalue;
  btree *curtree = FORMDATA;
  int i;  /* value is 1 to say override the default */
/*int i, value = 1;  |* value is 1 to say override the default */
  int ismultiple;

  /* generate data and datavalue from str */
  i = strchr (str, '=') - str;
  str[i] = '\0';
  data = strdup (str);
  str[i] = '=';
  datavalue = strdup (&str[i + 1]);

  /* first check our multiple selection handle
     This part is getting complicated and I don't
     really want to make it too complex :(
   */
  if (strcasecmp (multiple, "ON") == 0)
    {
      /* case study of the field names */
/*    if ((value = searchtree (&curtree, data)) != 0 || Value not referenced */
      if( searchtree (&curtree, data) != 0 ||
	  strcmp (MULTIPLE, data) == 0 ||
	  strcmp (EMPTYHANDLE, data) == 0 ||
	  strcmp (MULTIPLEMARK, data) == 0)
	ismultiple = 0;
      else
	ismultiple = 1;
    }
  else
    ismultiple = 0;

  /* then check if multiple selection ends */
  if (ismultiple == 0 &&
      (prevenv == NULL || strcmp (prevenv, data) != 0))
    {
      ismultiple = 0;  /* not multiple selection */
      if (prevenv)
	free (prevenv);
      prevenv = strdup (data);
    }
  else
    ismultiple = 1;  /* muliple selection */

  /* second, let's check if the string is empty */
  if (isemptyenv (datavalue))
    {
      /* do nothing if IGNORE */
      if (strcasecmp (emptyhandle, "IGNORE") == 0)
	{
	  free (data);
	  free (datavalue);
	  return 0;
	}
      if (strcasecmp (emptyhandle, "ERROR") == 0)
	{
	  emptyerror = data;
	  free (datavalue);
	  return 1;
	}
    }

  /* third, check predefined env */
  if (ismultiple == 0)
    {
      if (strstr (datavalue, "<!--"))  /* security problem */
	return -1;
      addtreeinput (curtree, data, datavalue);  /* optimize a bit */
      return 0;
    }
  else  /* combine the environmental variables */
    {
      int len;
      char *buff, *oldvalue = getnodedata (curtree);
      if (oldvalue == NULL)
	oldvalue = "";
      len = strlen (datavalue) + strlen (oldvalue) + strlen (multiplemark) + 1;
      buff = (char *)malloc (len);
      if (!buff)
	return -1;
      sprintf (buff, "%s%s%s", oldvalue, multiplemark, datavalue);
      if (strstr (buff, "<!--"))  /* security problem */
	return -1;
      addtreeinput (curtree, data, buff);
      free (data);
      free (datavalue);
      return 0;
    }
}

/* translate string and put fields into environment */
/* core routine */
int transf ()
{
  const char *c = inputstring;
  char  ch;
  int   i = 0;
  int   count = 0;
  char *str;
  int err = 0;

  startcgi ();  /* before making any translation, make a mark */

  str = (char *)malloc (inputlength + 1);
  if (!str)
    return -1;

  while (count <= inputlength)
    {
      ch = *c;
      switch (ch)
	{
	  case '\0':
	  case '&':
	    str[i] = '\0';
	    /* return values of handleenv are -1, 0, and 1.. */
	    /* -1 are not recoverable,
	       0 is okay.
	       1 is request for handle */
	    err = handleenv (str);
	    if (err > 0)
	      {
		free (str);
		return 1;
	      }
	    else if (err < 0)
	      {
		free (str);
		return -1;
	      }
	    i = 0;
	    c++;
	    count++;
	    continue;
	  case '%':
	    c++;
	    ch = hex2c (c);
	    str[i] = ch;
	    c++;
	    count += 2;
	    break;
	  case '+':
	    ch = ' ';
	    break;
	}
      str[i] = ch;
      i++;
      c++;
      count++;
    }

  free (str);
  return 0;
}

int getmethod ()
{
  int returnval;
  returnval = getinput ();
  if (returnval != 0)
    return returnval;
  returnval = transf ();
  if (freeinput == 1)
    {
      free (inputstring);
      freeinput = 0;
    }
  inputstring = NULL;
  return returnval;
}

btree *addtreeglobal (char* data, char *datavalue)
{
  btree *tmp = addtree (FORMDATA, data, datavalue, FORMDATA);
  if (FORMDATA == NULL)
    FORMDATA = tmp;
  tmp->linearnext = tmp;  /* let linearnext not to be a null */
  return tmp;
}

btree *addtreeinput (btree *tree, char *data, char *datavalue)
{
  btree *tmp = addtree (tree, data, datavalue, linearnext);
  if (linearnext == NULL)
    {
      linearnext = tmp;
      linearstart = tmp;
    }
  if (tmp->linearnext == NULL)
    linearnext = tmp;
  return tmp;
}
